{{
        Continuous Time Stamping on the propeller

        Jonathan "lonesock" Dummer

        Note: I measure 2x periods at a time, because
        I wanted rates above 2MHz, and could not sample
        faster than at about 1.5 MHz when capturing
        single periods.
}}

CON
  bufsize = 1024  ' we may not use all samples...during the calcs, if the integers are about to overflow it stops computiing
DAT
  buf   long  0[ bufsize + 1 ]
VAR
  byte  cog

PUB start( pin, timeout_clocks, ptr_to_sum_1_x_xx_y_yx )
  pin_mask := |<pin
  timeout := timeout_clocks
  buf_ptr_sub4 := @buf - 4                    
  return cog := cognew( @CTS_PASM_entry, ptr_to_sum_1_x_xx_y_yx ) + 1

PUB stop
  if cog
    cogstop( cog~ - 1 )


DAT
ORG 0
CTS_PASM_entry          ' Set up the variables
                        mov     ptr, buf_ptr_sub4
                        mov     i, buf_size
                        ' mark the time of the first edge
                        waitpne pin_mask, pin_mask
                        waitpeq pin_mask, pin_mask
                        mov     timein, cnt
                        
                        ' wait 1 more edge
                        waitpne pin_mask, pin_mask
                        waitpeq pin_mask, pin_mask
                        mov     tmp, #0   


get_samples             ' use this time to get some work done
                        add     ptr, #4
                        cmps    tmp, timeout wc         ' this test is one loop (2 periods) out of date
                        neg     tmp, timein             ' pre-adjust to the start time
                        
                        ' wait till low, then wait for a high edge
                        waitpne pin_mask, pin_mask                
                        waitpeq pin_mask, pin_mask
                        
                        ' grab the time, reference to the start time, store it
                        add     tmp, cnt                                                
                        wrlong  tmp, ptr

                        ' wait till low, then wait for a high edge
                        waitpne pin_mask, pin_mask                
                        waitpeq pin_mask, pin_mask                   

                        ' if we still have time and space, go back for another sample                                          
              if_c      djnz    i, #get_samples
               
                        ' done gathering, mark a final 0
                        add     ptr, #4 
                        wrlong  dirb, ptr

                        ' Process!
                        mov     sum1,  #0
                        mov     sumX,  #0                        
                        mov     sumXX, #0
                        mov     sumY,  #0 
                        mov     sumYX, #0
                        mov     sumDummy, #0
                        mov     i, #0
                        mov     ptr, buf_ptr_sub4
                        ' get each one back again
proc_loop               add     ptr, #4
                        rdlong  tmp, ptr wz
              if_z      jmp     #done_proc
                        add     i, #1
                        ' do the LSE accumulations
                        ' for period: numerator
                        mov     v1, i
                        mov     v2, tmp
                        call    #multu
                        ' first, check if we are about to overflow (a signed overflow...goes negative)
                        adds    sumDummy, vRes wc
              if_c      jmp     #done_proc
                        ' keep going, I guess
                        add     sumYX, vRes
                        ' for period: denominator
                        mov     v1, i
                        mov     v2, i
                        call    #multu
                        add     sumXX, vRes
                        ' the simple accumulations
                        add     sum1, #1
                        add     sumX, i
                        add     sumY, tmp
                        
                        ' more to do?
                        jmp     #proc_loop

done_proc               ' Store them?
                        rdlong  tmp, par wz
              if_z      jmp     #CTS_PASM_entry

                        ' yes                                             
                        wrlong  sum1, tmp
                        add     tmp, #4
                        wrlong  sumX, tmp
                        add     tmp, #4
                        wrlong  sumXX, tmp
                        add     tmp, #4
                        wrlong  sumY, tmp
                        add     tmp, #4
                        wrlong  sumYX, tmp

                        ' done
                        wrlong  dirb, par
                        ' just start over again
                        jmp     #CTS_PASM_entry


multu                   { shuffle the values so v1 has fewer high 1 bits
                        mov     vRes, v1
                        max     v1, v2
                        min     v2, vRes
                        '}
                        mov     vRes, #0
                        ' begin the core multiplication loop
:loop                   shr     v1, #1 wc, wz           ' get the lowest bit of v1
              if_c      add     vRes, v2                ' if it was a 1, add v2 to the accumulator
                        shl     v2, #1                  ' shift v2 left by 1 (to prep for the next v1 bit)
              if_nz     jmp     #:loop                  ' loop as long as v1 is not 0
multu_ret               ret

                
                        
pin_mask                long    0
buf_ptr_sub4            long    0
buf_size                long    bufsize
timeout                 long    0

' accumulators for the Least Squares Error calcs
sum1                    res     1
sumX                    res     1
sumY                    res     1
sumYX                   res     1
sumXX                   res     1

sumDummy                res     1 ' make sure we don't overflow the integers

' local vars
ptr                     res     1
i                       res     1
tmp                     res     1
timein                  res     1
v1                      res     1
v2                      res     1
vRes                    res     1
vTmp                    res     1