{{
        Test the Continuous Timestamping reciprocal frequency counter

        Jonathan "lonesock" Dummer
}}
CON
  _CLKMODE = XTAL1 + PLL16X
  _XINFREQ = 5_000_000

  ' serial pins
  term_RX       = 31
  term_TX       = 30
  term_bps      = 460_800

  test_pin      = 4


OBJ
  term:         "FFDS1"
  fc:           "CTS_FreqCounter"
  f32:          "F32"

dat
  spinidx byte 0
  spinchars byte "|/-\"

PRI frac32( num, den )
  repeat 32
    result += result
    if num => den
      result += 1
      num -= den
    num += num

PUB Main | counts, Pdrive, fdrive, fs, fnum, fden, {
        } ptr_to_sums, sum1, sumX, sumXX, sumY, sumYX

  ' set up counter A to do a NCO squarewave out
  ctra := constant( %00100 << 26 )     ' PLL output, single ended
  ctra |= test_pin                     ' output pin
  ' set a frequency (CTS seems to handle just up to 2.35 MHz)
  fs := 100
  ' The equation would look something like:
  '   frqa = fs * (1 << 32) / clkfreq
  ' But that is numerically horrible to do in integer math
  frqa := frac32( fs * 2, clkfreq )
  dira[ test_pin ] := true

  ' start up the frequency counter
  ptr_to_sums := @sum1
  fc.start( test_pin, clkfreq / 100, @ptr_to_sums )

  ' and I need floating point (to make my life easier [8^)
  f32.start
    
  ' Start the serial link
  if ina[term_RX]
    ' the RX pin is high, so there is probably power to the FTDI chip
    term.start( term_RX, term_TX, term_bps )
  else
    ' just RX, since the FTDI is probably unpowered
    term.start( term_RX, -1, term_bps )  

  '{ START SWEEP
  waitcnt( clkfreq * 2 + cnt )
  term.tx( 0 )

  ' limit at about 2_353_500
  repeat fdrive from 2_352_000 to 2_354_000 step 10
    'repeat fdrive from 1_000_000 to 2_000_000 step 1000
    ' set the frequency
    frqa := frac32( fdrive * 2, clkfreq )
    ' wait a bit
    waitcnt( clkfreq / 100 + cnt )
    ' signal that I want a new reading
    ptr_to_sums := @sum1
    repeat
    while ptr_to_sums
    ' we got one, now convert the integers to floats
    sum1  := f32.FFloat( sum1 )
    sumX  := f32.FFloat( sumX )
    sumXX := f32.FFloat( sumXX )
    sumY  := f32.FFloat( sumY )
    sumYX := f32.FFloat( sumYX )
    ' and do the math to calculate the 2x period
    'fs := f32.FDiv( sumYX, sumXX ) { simple gain-only equation
    fnum := f32.FSub( f32.FMul( sum1, sumYX ), f32.FMul( sumX, sumY ) )
    fden := f32.FSub( f32.FMul( sum1, sumXX ), f32.FMul( sumX, sumX ) )
    fs := f32.FDiv( fnum, fden )
    ' The full version } 

    ' convert 2x period to frequency
    fs := f32.FDiv( f32.FFloat( clkfreq * 2 ), fs )
    ' result back as an integer
    fs := f32.FRound( fs )

    ' send back results
    term.tx( 13 )
    term.dec( fdrive )
    term.tx( 9 )
    term.dec( fs )
    term.tx( 9 )
    term.dec( fdrive - fs )

  'waitpne( 0,0,0 )
  term.rx
  ' END SWEEP }


  '{ set up counter A to do a PLL squarewave out, trying for NOT synced to the system clock
  ctra := constant( %00010 << 26 )     ' PLL output, single ended
  ctra |= constant( %000 << 23 )       ' VCO / 128 (and VCO is 16x the NCO frequency, so NCO / 8)
  ctra |= test_pin                     ' output pin
  ' set a frequency (docs say 4MHz to 8MHz to be stable, limit seems right around 14MHz)
  fs := 6_000_000 ' 4_000_000 to 8_000_000
  ' The equation would look something like:
  '   frqa = fs * (1 << 32) / clkfreq
  ' But that is numerically horrible to do in integer math
  frqa := frac32( fs * 2, clkfreq )
  ' OR...just go all the way
  frqa := posx ' WAY outside the stable range for the PLL!!   
  dira[ test_pin ] := true
  '}


  ' monitor period
  repeat
    waitcnt( clkfreq / 4 + cnt )
    ' signal that I want a new reading
    ptr_to_sums := @sum1
    repeat
    while ptr_to_sums
    ' we got one, now convert the integers to floats
    sum1  := f32.FFloat( sum1 )
    sumX  := f32.FFloat( sumX )
    sumXX := f32.FFloat( sumXX )
    sumY  := f32.FFloat( sumY )
    sumYX := f32.FFloat( sumYX )
    ' and do the math to calculate the period
    'fs := f32.FDiv( sumYX, sumXX ) { simple gain-only equation
    fnum := f32.FSub( f32.FMul( sum1, sumYX ), f32.FMul( sumX, sumY ) )
    fden := f32.FSub( f32.FMul( sum1, sumXX ), f32.FMul( sumX, sumX ) )
    fs := f32.FDiv( fnum, fden )
    ' The full version } 

    ' convert period to frequency
    fs := f32.FDiv( f32.FFloat( clkfreq * 2 ), fs )
    ' result back as an integer
    fs := f32.FRound( fs )
    term.tx( 0 )
    term.tx( spinchars[ ++spinidx & 3 ] )
    term.tx( " " )
    term.dec( fs )
